// Copyright 2017 The Johns Hopkins University Applied Physics Laboratory.
// Licensed under the MIT License. See LICENSE.txt in the project root for full license information.


#pragma once

#include <vector>
#include <utility>
#include <iterator>

#include <eopcc/utility.hpp>
#include <eopcc/cloud.hpp>

namespace eopcc {

template<typename T>
class GridModel {
public:
  typedef std::pair<double, double> spacing_t;
private:
    std::string wkt_info_;

    spacing_t grid_spacing_;
    int x_dim_, y_dim_;
    bool valid_extent_;
    Extent extent_;

    T invalid_value_;

    void setupGrid(int x_dim, int y_dim, std::pair<double, double> grid_spacing) {
        grid.resize(x_dim*y_dim);
        grid_spacing_ = grid_spacing;
        x_dim_ = x_dim;
        y_dim_ = y_dim;
        invalid_value_ = std::numeric_limits<T>::lowest();
        setAll(invalid_value_);

    }
public:
    std::vector<T> grid;


    GridModel() : grid_spacing_(0,0), x_dim_(0), y_dim_(0), valid_extent_(false) {

    }

    GridModel(const Extent& extent, std::pair<double, double> grid_spacing) {
        assert(std::get<0>(grid_spacing) > 0.0);
        assert(std::get<1>(grid_spacing) > 0.0);

        extent_ = extent;
        valid_extent_ = true;

        int x_dim = (extent.max_x - extent.min_x)/std::get<0>(grid_spacing) + 1;
        int y_dim = (extent.max_y - extent.min_y)/std::get<1>(grid_spacing) + 1;

        setupGrid(x_dim, y_dim, grid_spacing);
    }
    GridModel(int x_dim, int y_dim, std::pair<double, double> grid_spacing) {
        setupGrid(x_dim, y_dim, grid_spacing);
    }

    const std::pair<double, double>& spacing() const { return grid_spacing_; }
    double spacingX() const { return std::get<0>(grid_spacing_); }
    double spacingY() const { return std::get<1>(grid_spacing_); }
    int dimX() const { return x_dim_; }
    int dimY() const { return y_dim_; }

    void SetInvalidValue(const T& value) { invalid_value_ = value; }
    T GetInvalidValue() const { return invalid_value_; }

    void SetExtent(Extent& extent) {
        extent_ = extent;
        valid_extent_ = true;
    }

    void SetWKT(const std::string& info) {
        wkt_info_ = info;
    }

    std::string GetWKT() const {
        return wkt_info_;
    }

    template <typename H> GridModel<H> convertTo() const {
        GridModel<H> grid(extent_, grid_spacing_);
        grid.SetWKT(wkt_info_);

        for (int y=0; y < y_dim_; ++y) {
            for (int x=0; x < x_dim_; ++x) {
                if (operator()(x,y) != invalid_value_) {
                    grid(x,y) = (H) operator()(x,y);
                }
            }
        }

        return grid;
    }

    const T& at(double x, double y) const {
        assert(valid_extent_);
        assert(x >= extent_.min_x);
        assert(y <= extent_.max_y);

        return operator()((x - extent_.min_x)/spacingX(), (extent_.max_y - y)/spacingY());
    }

    T& at(double x, double y) {
        assert(valid_extent_);
        assert(x >= extent_.min_x);
        assert(y <= extent_.max_y);

        return operator()((x - extent_.min_x)/spacingX(), (extent_.max_y - y)/spacingY());
    }

    T& operator() (int x, int y) {
        return grid[x + y*x_dim_];
    }
    const T& operator() (int x, int y) const {
        return grid[x + y*x_dim_];
    }

    typedef std::pair<int, int> location_t;

    location_t location_of(double x, double y) const {
        return location_t((x - extent_.min_x)/spacingX(), (extent_.max_y - y)/spacingY());   
    }

    T& operator() (location_t grid_loc) {
        return operator()(std::get<0>(grid_loc), std::get<1>(grid_loc));
    }

    const T& operator() (location_t grid_loc) const {
        return operator()(std::get<0>(grid_loc), std::get<1>(grid_loc));
    }


    bool valid_location(int x, int y) const {
        if (x < 0 || x >= dimX())
            return false;
        if (y < 0 || y >= dimY())
            return false;

        return true;
    }

    bool is_inside(double x, double y) const {
        return in_range(x, extent_.min_x, extent_.max_x) &&
               in_range(y, extent_.min_y, extent_.max_y); 
    }

    bool contains(double x, double y) const  {
        return is_inside(x,y) && at(x,y) != invalid_value_;
    }

    bool valid_location(location_t grid_loc) const {
        if (std::get<0>(grid_loc) < 0 || std::get<0>(grid_loc) >= dimX())
            return false;
        if (std::get<1>(grid_loc) < 0 || std::get<1>(grid_loc) >= dimY())
            return false;


        return true;
    }

    bool patchHole(int x, int y, std::pair<int, int> window_size, const GridModel<int>& counts) {
        long long x_start = x - std::get<0>(window_size)/2;
        long long y_start = y - std::get<1>(window_size)/2;

        T sum = 0.0;
        int count = 0;
        for (long long y_pos=0; y_pos < std::get<1>(window_size); ++y_pos) {
            for (long long x_pos=0; x_pos < std::get<0>(window_size); ++x_pos) {
                if (in_range<long long>(x_start + x_pos, 0, dimX())
                        && in_range<long long>(y_start + y_pos, 0, dimY())
                        && counts(x_start + x_pos, y_start + y_pos)) {

                    sum += this->operator()(x_start + x_pos, y_start + y_pos);
                    count += 1;
                }
            }
        }
        if (count) {
            this->operator()(x,y) = sum / count;
            return true;
        } else {
            return false;
        }
    }

    Extent extent() const {
        return extent_;
    }

    void setAll(T value) {
        grid.assign(grid.size(), value);
    }



    void overlap(const Extent& other, location_t& top_left, location_t& bottom_right) const {
        top_left = location_of(other.min_x, other.max_y);
        bottom_right =location_of(other.max_x, other.min_y);

        std::get<0>(top_left) = std::max(std::get<0>(top_left), 0);
        std::get<1>(top_left) = std::max(std::get<1>(top_left), 0);
        std::get<0>(bottom_right) = std::min(std::get<0>(bottom_right), dimX());
        std::get<1>(bottom_right) = std::min(std::get<1>(bottom_right), dimY());
    }
};




void MaxHeightGrid(GridModel<double>& max_height_grid, const PointCloud& point_cloud, Point shift);
GridModel<double> MaxHeightGrid(const PointCloud& point_cloud, const GridModel<double>::spacing_t& grid_spacing);
GridModel<double> AverageHeightGrid(const PointCloud& point_cloud, const GridModel<double>::spacing_t& grid_spacing);


} // namespace eopcc